/****************************/
/*  Join several S19 files  */
/*   to one file            */
/****************************/

/*************************************************************************
*               R E V I S I O N   H I S T O R Y                       
*                         Release 0
*
* Rev #    Date        Modifications/Corrections for new Revision
* -----  ---------     ---------------------------------------------------
*  0.01    ------      ---
*
*  0.02    170210   1. param -ain,out mod: both in and out optional
*                   2. param -az added: short form of -a,0
*
*
*
*
*/
char *version = "0.02  2017-02-10";

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#define MAXBUFLEN   0x10000
#define INFILESMAX  32
#define S19LEN      32

enum {
    DATA_RECORD = 0,
    INIT_RECORD,
    END_RECORD
};

typedef unsigned int uint;
typedef unsigned char uchar;

typedef struct infile {
    char *name;
    int inaddr;
    int outaddr;

} INFILE;

int debugLevel = 0;

FILE *fe;
char *cmdbase;
char *cmd;

int rlength = S19LEN;

char outfile[250];
int pattern = 0xff;
int noS0 = 0;

INFILE infiles[INFILESMAX];
int infilescnt = 0;

int outputall = 0;
int outsize = MAXBUFLEN;
int outstart = 0;
uchar buf[MAXBUFLEN+128];
char mark[MAXBUFLEN+128];
char s0name[64];

int outbytecnt;

void init()
{
   int i;

   for (i = 0; i < outsize; i++) {
      buf[i] = pattern;
      mark[i] = 0;
   }

}

void prntbuf(FILE *fh, int type, int addr, uchar *buf, int count)
{
    int i;
    int checksum;
    char tt[200];
    char *d;

    switch (type) {
    case INIT_RECORD:
        strcpy(tt, "S0");
        break;
    case DATA_RECORD:
        strcpy(tt, "S1");
        outbytecnt += count;
        break;
    case END_RECORD:
        strcpy(tt, "S9");
        break;
    }

    d = tt + 2;

    count += 3;		/* include address and csum in count */
    sprintf(d, "%02X%04X", count,  addr);
    d += strlen(d);

    checksum = ((addr >> 8) & 0xff) + (addr & 0xff) + (count & 0xff);
    count -= 3;

    for (i = 0; i < count; i++) {
        checksum += buf[i] & 0xff;
        sprintf(d, "%02X", buf[i] & 0xff);
        d += 2;
    }

    checksum = (~checksum) & 0xff;

    sprintf(d, "%02X", checksum);
    fprintf(fh, "%s\n", tt);
}

/*
 *      convert ascii string (of hex digits) to binary values
 */
int readhex(char *p, int sz, char **pp)
{
    int c;
    int val;

    val = 0;
    while (sz) {
        c = *p;

        if (c >= 'A' && c <= 'F') {
            c = c - 'A' + 10;
        }
        else if (c >= 'a' && c <= 'f') {
            c = c - 'a' + 10;
        }
        else if (c >= '0' && c <= '9') {
            c = c - '0';
        }
        else {
            break;
        }

        p++;
        val = (val * 16) + c;
        --sz;
    }

    if (pp) {
        *pp = p;
    }
    return val;
}

/*
 *
 * S0 0F 0000 6C69676874706F772E61736D 13
 * S1 23 F800 0F8E00FF860F970086FF970686FF970097044F9705D602C4F054D78286009708 FA
 * S9 03 0000 FC
 *
 * Return: type or -1 if error
 */
int splitline(uchar *buf, int *lenp, int *addrp, char *p)
{
    int type;
    int addr;
    int count;
    int val;
    int cs;
    int ind;

// line header
    if (toupper(*p) != 'S') {
        fprintf(fe, "no S\n");
        return -1;
    }
    p++;

    if (*p == '0') {
        type = INIT_RECORD;
    }
    else if (*p == '1') {
        type = DATA_RECORD;
    }
    else if (*p == '9') {
        type = END_RECORD;
    }
    else {
        fprintf(fe, "no 0, 1, 9\n");
        return -1;
    }
    p++;

    cs = 0;

// length
    if (!isxdigit(*p)) {
        fprintf(fe, "no length\n");
        return -1;
    }
    count = readhex(p, 2, &p);
    cs += count;

    count -= 3;
// addr
    if (!isxdigit(*p)) {
        fprintf(fe, "no addr\n");
        return -1;
    }
    addr = readhex(p, 4, &p);
    cs += ((addr >> 8) & 0xff) + (addr & 0xff);

//    fprintf(fe, "addr: %04X, len: %02X\n", addr, count);

// data
    ind = 0;
    while (ind < count) {
        if (!isxdigit(*p)) {
            fprintf(fe, "data missing\n");
            return -1;
        }
        val = readhex(p, 2, &p);
        cs += val;
        buf[ind] = val;
        ind++;
    }

// checksum
    cs = (~cs) & 0xff;

    if (!isxdigit(*p)) {
        fprintf(fe, "no CS\n");
        return -1;
    }
    val = readhex(p, 2, &p);
    if (val != cs) {
        fprintf(fe, "CS error\n");
        return -1;
    }

    *lenp = count;
    *addrp = addr;
    return type;
}

int ReadLine(FILE *ff, char *txtbuf, int buflen)
{
    int i, c;
    int rv;

    --buflen; // make space for endmark

    rv = 0;
    i = 0;
    while ((c = fgetc(ff)) != EOF && c != '\n') {
        if (c == 0x1a) {    // skip MSDOS END-OF-FILE char
        }
        else if (c == '\r') {    // skip CR
        }
        else if (i < buflen) {
            txtbuf[i++] = c;
        }
    }

    txtbuf[i] = 0;

    if (c == EOF) {
        return EOF;
    }
    return rv;
}

void readinput(INFILE *fp)
{
    FILE *fh;
    char txt[200];
    int eof;
    char *p;
    int len;
    int addr;
    int ind;
    int type;
    uchar dta[128];
    int line;
    int offset;
    int offsetValid = 0;
    int i;

    fh = fopen(fp->name, "r");
    if (fh == NULL) {
        fprintf(fe, "Infile not found: '%s'\n", fp->name);
        exit(1);
    }

    if (fp->inaddr >= 0) {
        if (fp->outaddr >= 0) {
            offset = fp->outaddr - fp->inaddr; // in and out given
        }
        else {
            offset = fp->inaddr;    // in only
        }
        offsetValid = 1;
    }
    else if (fp->outaddr < 0) { // none given
        offset = 0;
        offsetValid = 1;
    }

    if (debugLevel & 1) {
        fprintf(fe, "offs: %04X: out %04X: in %04X\n", offset, fp->outaddr, fp->inaddr);
        fprintf(fe, "size: %d, start: %04X\n", outsize, outstart);
    }

    line = 0;
    eof = 0;
    while (eof != EOF) {
        eof = ReadLine(fh, txt, 200);
        line++;

        if (txt[0] == 0) {  // empty line
        }
        else {
            type = splitline(dta, &len, &addr, txt);
            if (type < 0) {
                fprintf(fe, "%s: S-record error at line %d\n", fp->name, line);
                exit(1);
            }
            else if (type == DATA_RECORD) {
                if (offsetValid == 0) {         // set offset with help of first data record
                    offset = fp->outaddr - addr;
                    offsetValid = 1;
                }

                ind = addr + offset;
                if (ind < outstart || ind >= outstart + outsize) {
                    fprintf(fe, "%s: Warning: S-record at line %d out of address range 0x%04X (addr %04X)\n",
                            fp->name, line, ind, addr);
                }
                else {
                    for (i = 0; i < len; i++) {
                        buf[ind+i] = dta[i];
                        mark[ind+i] = 1;
                    }
                }
            }
            else if (type == INIT_RECORD) {
                if (s0name[0] == 0) {
                    dta[len] = 0;
                    strcpy(s0name, dta);
                }
            }
        }
    }

    fclose(fh);
}

void writeoutput()
{
    FILE *fh;
    int ftype;
    int addr;
    int len;
    int ind;
    int stop;
    int i;

    if (outfile[0]) {
        fh = fopen(outfile, "w");
        if (fh == NULL) {
            fprintf(fe, "Cannot open outfile: '%s'\n", outfile);
            exit(1);
        }
        ftype = 1;
    }
    else {
        fh = stdout;
        ftype = 0;
    }

    if (debugLevel & 1) {
        fprintf(fe, "Write to outfile: '%s'\n", outfile);
    }

    if (noS0 == 0) {
        len = strlen(s0name);
        prntbuf(fh, INIT_RECORD, 0, s0name, len);
    }

    outbytecnt = 0;

    if (outputall) {
        ind = outstart;
        stop = outstart + outsize;

        while (ind < stop) {
            len = stop - ind;
            if (len > rlength) {
                len = rlength;
            }
            prntbuf(fh, DATA_RECORD, ind, buf+ind, len);
            ind += len;
        }
    }
    else {
        ind = outstart;
        stop = outstart + outsize;

        while (ind < stop) {
            if (mark[ind]) {
                len = stop - ind;
                if (len > rlength) {
                    len = rlength;
                }

                for (i = 0; i < len; i++) {
                    if (mark[ind+i] == 0) {
                        break;
                    }
                }
                if (len > i) {
                    len = i;
                }

                prntbuf(fh, DATA_RECORD, ind, buf+ind, len);
                ind += len;
            }
            else {
                ind++;
            }
        }
    }

    prntbuf(fh, END_RECORD, 0, buf, 0);

    if (ftype == 1) {
        fclose(fh);
    }

    fprintf(fe, "Data bytes written to output: %d\n", outbytecnt);
}

void parsesize(char *p)
{
    int val1;
    int mult;

    mult = 1;
    val1 = strtol(p, &p, 0);
    if (*p == 'k') {
        mult = 1024;
        p++;
    }
    outsize = val1 * mult;
    if (*p == ',') {
        p++;
        mult = 1;
        val1 = strtol(p, &p, 16);
        if (*p == 'k') {
            mult = 1024;
            p++;
        }
        outstart = val1 * mult;
    }

}

void usage(char *arg)
{
    int cmdlen = strlen(cmdbase);
    int i;

    printf("\n");
    printf("usage:  %s [-v] [-h] [-f fill] [-l length] [-n name] [-o outfile]\n",
           cmdbase);
    printf("usage:  %*s [-s size[,start]] [-ain,out] file [-ain,out] file ... \n",
           cmdlen, "");
    printf("\n");
    printf("  -v           print version\n");
    printf("  -h           print help\n");
    printf("  -f fill      fill pattern, in hex (defalts to ff)\n");
    printf("  -l length    length of records in output, decimal (def: 32)\n");
    printf("  -o outfile   output file\n");
    printf("  -n-          skip S0 record in output file\n");
    printf("  -n name      string to place in the S0 header, first input taken if not given\n");
    printf("  -s size[,start] size and start of output area, in bytes, or kB append a 'k'\n");
    printf("                  size in decimal or hex with prefix 0x. Start always in hex\n");
    printf("  -a[in][,out] set address translation. addr in hex, no translation if omitted\n");
    printf("                 in: input-addr (take address from first record if omitted) \n");
    printf("                 out: output-addr\n");
    printf("  -az          short form of -a,0\n");
    printf("  file         input files\n");
    printf("               repeat param -a for each infile\n");
    printf("\n");

}

enum {
    NXT_OUTFILE = 1,
    NXT_SIZE,
    NXT_FILL,
    NXT_NAME,
    NXT_LENGTH,
};

int GetParameter(int argc, char *argv[])
{
    int i;
    int nxtarg;
    char *p;
    int val1 = -1;
    int val2 = -1;
    int mult;

    cmd = argv[0];
    cmdbase = strrchr(cmd, '/');
    cmdbase = cmdbase ? cmdbase+1 : cmd;

    nxtarg = 0;
    for (i = 1; i < argc; i++) {
        if (*argv[i] == '-') {
            nxtarg = 0;
            switch (argv[i][1]) {
            case 'v':   /* print version */
                printf("\n%s:   version %s\n\n", cmd, version);
                exit(0);

            default:
                fprintf(fe, "Unknown option: %s\n", argv[i]);
                usage(argv[i]);
                exit(2);

            case 'h':   /* help */
                usage(argv[i]);
                exit(0);

            case '-':
                break;

            case 'D':   /* help */
                p = argv[i]+2;
                if (isxdigit(*p)) {
                    debugLevel = strtol(p, &p, 16);
                }
                else {
                    debugLevel = 1;
                }
                break;

            case 'a':   /* addr-offs */
                p = argv[i]+2;
                if (*p == 'z') {
                    val2 = 0;
                }
                else {
                    if (isxdigit(*p)) {
                        val1 = strtol(p, &p, 16);
                    }
                    if (*p == ',') {
                        p++;
                        if (isxdigit(*p)) {
                            val2 = strtol(p, &p, 16);
                        }
                    }
                }
                break;

            case 'f':   /* fill pattern */
                p = argv[i]+2;
                outputall = 1;
                if (*p) {
                    if (isxdigit(*p)) {
                        pattern = strtol(p, &p, 16);
                    }
                }
                else {
                    nxtarg = NXT_FILL;
                }
                break;

            case 'l':   /* record length */
                p = argv[i]+2;
                if (*p) {
                    if (isdigit(*p)) {
                        rlength = strtol(p, &p, 10);
                    }
                }
                else {
                    nxtarg = NXT_LENGTH;
                }
                break;

            case 'o':   /* outfile */
                p = argv[i]+2;
                if (*p) {
                    strcpy(outfile, p);
                }
                else {
                    nxtarg = NXT_OUTFILE;
                }
                break;

            case 'n':   /* s0name */
                p = argv[i]+2;
                if (!strcmp(p, "-")) {
                    noS0 = 1;
                }
                else if (*p) {
                    strcpy(s0name, p);
                }
                else {
                    nxtarg = NXT_NAME;
                }
                break;

            case 'S':   /* size of output area */
                p = argv[i]+2;
                if (*p == '0') {
                    noS0 = 1;
                }
                break;

            case 's':   /* size of output area */
                p = argv[i]+2;
                if (*p) {
                    parsesize(p);
                }
                else {
                    nxtarg = NXT_SIZE;
                }
                break;

            }
        }
        else {
            p = argv[i];
            switch (nxtarg) {
            case 0:
                if (infilescnt < INFILESMAX) {
                    infiles[infilescnt].name = p;
                    infiles[infilescnt].inaddr = val1;
                    infiles[infilescnt].outaddr = val2;
                    infilescnt++;
                    val1 = -1;
                    val2 = -1;
                }
                break;

            case NXT_OUTFILE:
                strcpy(outfile, p);
                nxtarg = 0;
                break;

            case NXT_NAME:
                strcpy(s0name, p);
                nxtarg = 0;
                break;

            case NXT_FILL:
                if (isxdigit(*p)) {
                    pattern = strtol(p, &p, 16);
                }
                nxtarg = 0;
                break;

            case NXT_LENGTH:
                if (isdigit(*p)) {
                    rlength = strtol(p, &p, 10);
                }
                nxtarg = 0;
                break;

            case NXT_SIZE:
                parsesize(p);
                nxtarg = 0;
                break;

            default:
                nxtarg = 0;
                break;
            }
        }
    }
    return 0;
}

int main(int argc, char *argv[])
{
    int i;

    fe = stderr;

    GetParameter(argc, argv);

    init();

    for (i = 0; i < infilescnt; i++) {
        readinput(infiles+i);
    }

    writeoutput();

    exit(0);
}

/* end */

